/*
//
// Copyright (C) 2006, 2007, 2008, 2009 Jean-Franois DEL NERO
//
// This file is part of HxCFloppyEmulator.
//
// HxCFloppyEmulator may be used and distributed without restriction provided
// that this copyright statement is not removed from the file and that any
// derivative work contains the original copyright notice and the associated
// disclaimer.
//
// HxCFloppyEmulator is free software; you can redistribute it
// and/or modify  it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// HxCFloppyEmulator is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with HxCFloppyEmulator; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//
*/
///////////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------------------------------//
//-------------------------------------------------------------------------------//
//-----------H----H--X----X-----CCCCC----22222----0000-----0000------11----------//
//----------H----H----X-X-----C--------------2---0----0---0----0--1--1-----------//
//---------HHHHHH-----X------C----------22222---0----0---0----0-----1------------//
//--------H----H----X--X----C----------2-------0----0---0----0-----1-------------//
//-------H----H---X-----X---CCCCC-----222222----0000-----0000----1111------------//
//-------------------------------------------------------------------------------//
//----------------------------------------------------- http://hxc2001.free.fr --//
///////////////////////////////////////////////////////////////////////////////////
// File : snes_smc_DiskFile.c
// Contains: SMC rom floppy image loader and plugins interfaces
//
// Written by:	DEL NERO Jean Francois
//
// Change History (most recent first):
///////////////////////////////////////////////////////////////////////////////////

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "hxc_floppy_emulator.h"
#include "internal_floppy.h"
#include "floppy_loader.h"
#include "floppy_utils.h"

#include "../common/crc.h"
#include "../common/iso_ibm_track.h"

#include "snes_smc_loader.h"

#include "os_api.h"

#include "../fat12floppy_loader/fat12.h"

extern unsigned char msdos_bootsector;

int snes_smc_libIsValidDiskFile(HXCFLOPPYEMULATOR* floppycontext,char * imgfile)
{
	FILE * f;
	unsigned char * fileheader;
	char * filepath;
	int fileok;
	
	fileok=1;
	floppycontext->hxc_printf(MSG_DEBUG,"snes_smc_libIsValidDiskFile %s",imgfile);
	if(imgfile)
	{
		
		f=fopen(imgfile,"r+b");
		if(f)
		{
			fileheader=(unsigned char*)malloc(512);
			if(fileheader)
			{
				fread(fileheader,1,512,f);
				if( fileheader[8]==0xaa && fileheader[9]==0xbb )
				{
					
					switch(fileheader[10])
					{
					case 1:
						floppycontext->hxc_printf(MSG_INFO_1,"File type : Super Magic Card saver data");
						break;
					case 2:
						floppycontext->hxc_printf(MSG_INFO_1,"File type : Magic Griffin program (PC-Engine)");
						break;
					case 3:
						floppycontext->hxc_printf(MSG_INFO_1,"File type : Magic Griffin SRAM data");
						break;
					case 4:
						floppycontext->hxc_printf(MSG_INFO_1,"File type : SNES program");
						break;
					case 5:
						floppycontext->hxc_printf(MSG_INFO_1,"File type : SWC & SMC password, SRAM data");
						break;
					case 6:
						floppycontext->hxc_printf(MSG_INFO_1,"File type : Mega Drive program");
						break;
					case 7:
						floppycontext->hxc_printf(MSG_INFO_1,"File type : SMD SRAM data");
						break;
					case 8:
						floppycontext->hxc_printf(MSG_INFO_1,"File type : SWC & SMC saver data");
						break;
					default:
						fileok=0;					
						break;
					}
					
					if(fileok)
					{
						floppycontext->hxc_printf(MSG_DEBUG,"SMC/SMD file !");
						free(fileheader);
						fclose(f);
						return LOADER_ISVALID;
					}
					else
					{
						floppycontext->hxc_printf(MSG_DEBUG,"non SMC/SMD file !");
						free(fileheader);
						fclose(f);
						return LOADER_BADFILE;
					}
					
				}
				else
				{
						if(!strncmp(&fileheader[8],"SUPERUFO",8))
						{
							floppycontext->hxc_printf(MSG_INFO_1,"File type : SUPERUFO SMC");
						}
						else
						{
							filepath=malloc(strlen(imgfile)+1);
							if(filepath!=0)
							{
								sprintf(filepath,"%s",imgfile);
								strlwr(filepath);
								if(!strstr( filepath,".smc" ))
								{
									fileok=0;
								}
								else
								{
									floppycontext->hxc_printf(MSG_INFO_1,"File type : Super Pro Fighter SMC?");
								}

								free(filepath);
							}

							if(!fileok) floppycontext->hxc_printf(MSG_INFO_1,"unknow file type !");
				
						}

					if(fileok)
					{
						floppycontext->hxc_printf(MSG_DEBUG,"SMC/SMD file !");
						free(fileheader);
						fclose(f);
						return LOADER_ISVALID;
					}
					else
					{
						floppycontext->hxc_printf(MSG_DEBUG,"non SMC/SMD file !");
						free(fileheader);
						fclose(f);
						return LOADER_BADFILE;
					}
				}
				
			}
			fclose(f);
			return LOADER_INTERNALERROR;
		}
		return LOADER_ACCESSERROR;
		
	}
	
	return LOADER_BADPARAMETER;
}



int snes_smc_libLoad_DiskFile(HXCFLOPPYEMULATOR* floppycontext,FLOPPY * floppydisk,char * imgfile,void * parameters)
{
	unsigned int i,j;
	unsigned int file_offset;
	char* trackdata;
	unsigned char * flatimg;
	int tracklen;
	int gap3len,interleave,rpm;
	int sectorsize;
	int numberofcluster;
	unsigned long   fatposition;
	unsigned long   rootposition;
	unsigned long   dataposition;
	int pcbootsector;
	int dksize;
	CYLINDER* currentcylinder;
	SIDE* currentside;



	FATCONFIG fatconfig;

	floppycontext->hxc_printf(MSG_DEBUG,"snes_smc_libLoad_DiskFile %s",imgfile);
	
	floppydisk->floppyNumberOfTrack=126;
	floppydisk->floppyNumberOfSide=2;
	floppydisk->floppySectorPerTrack=18;
	floppydisk->floppyBitRate=500000;
	rpm=300;
	pcbootsector=1;

	dksize=floppydisk->floppyNumberOfTrack*
			(floppydisk->floppySectorPerTrack*floppydisk->floppyNumberOfSide*512);


	floppycontext->hxc_printf(MSG_INFO_1,"floppy size:%dkB, %d tracks, %d side(s), %d sectors/track, rpm:%d",dksize/1024,floppydisk->floppyNumberOfTrack,floppydisk->floppyNumberOfSide,floppydisk->floppySectorPerTrack,rpm);
	
	flatimg=(char*)malloc(dksize);
	if(flatimg!=NULL)
	{
		memset(flatimg,0,dksize);
		if(pcbootsector)
			memcpy(flatimg,&msdos_bootsector,512);

		fatconfig.sectorsize=512;
		fatconfig.clustersize=2;
		fatconfig.reservedsector=1;
		fatconfig.numberoffat=2;
		fatconfig.numberofrootentries=112;// 7secteurs = 32 *512
		fatconfig.nbofsector=(floppydisk->floppyNumberOfTrack*
								(floppydisk->floppySectorPerTrack*floppydisk->floppyNumberOfSide));
		fatconfig.nbofsectorperfat=((fatconfig.nbofsector-(fatconfig.reservedsector+(fatconfig.numberofrootentries/32)))/((fatconfig.sectorsize*8)/12))+1;
		//sprintf(&flatimg[CHSTOADR(0,0,0)+3],"HXC.EMU");

		*( (unsigned short*) &flatimg[0x0B])=fatconfig.sectorsize; //Nombre d'octets par secteur
		*( (unsigned char*)  &flatimg[0x0D])=fatconfig.clustersize; //Nombre de secteurs par cluster (1, 2, 4, 8, 16, 32, 64 ou 128).
		*( (unsigned short*) &flatimg[0x0E])=fatconfig.reservedsector; //Nombre de secteur rserv en comptant le secteur de boot (32 par dfaut pour FAT32, 1 par dfaut pour FAT12/16).
		*( (unsigned char*)  &flatimg[0x10])=fatconfig.numberoffat; //Nombre de FATs sur le disque (2 par dfaut).
		*( (unsigned short*) &flatimg[0x11])=fatconfig.numberofrootentries; //Taille du rpertoire racine (0 par dfaut pour FAT32).
		*( (unsigned short*) &flatimg[0x13])=fatconfig.nbofsector; //Nombre total de secteur 16-bit (0 par dfaut pour FAT32).

		if(	floppydisk->floppyBitRate==250000)
		{
			*( (unsigned char*)  &flatimg[0x15])=0xF9; //Type de disque (0xF8 pour les disques durs, 0xF0 pour les disquettes). 0xF9 Double sided, 80 tracks per side, 9 sectors per track 
		}
		else
		{
			*( (unsigned char*)  &flatimg[0x15])=0xF0; //Type de disque (0xF8 pour les disques durs, 0xF0 pour les disquettes). 0xF9 Double sided, 80 tracks per side, 9 sectors per track 
		}

		*( (unsigned short*) &flatimg[0x16])=fatconfig.nbofsectorperfat; //Taille d'une FAT en secteurs (0 par dfaut pour FAT32).
		*( (unsigned short*) &flatimg[0x18])=floppydisk->floppySectorPerTrack; //Sectors per track
		*( (unsigned short*) &flatimg[0x1a])=floppydisk->floppyNumberOfSide; //Number of heads.
		*( (unsigned short*) &flatimg[0x1FE])=0xAA55;//End of sector marker (0x55 0xAA)		

		fatposition=fatconfig.sectorsize*fatconfig.reservedsector;
		rootposition=((fatconfig.reservedsector)+(fatconfig.numberoffat*fatconfig.nbofsectorperfat))*fatconfig.sectorsize;
		dataposition=(((fatconfig.reservedsector)+(fatconfig.numberoffat*fatconfig.nbofsectorperfat))*fatconfig.sectorsize)+(32*fatconfig.numberofrootentries);		
		
		numberofcluster=(fatconfig.nbofsector-(dataposition/fatconfig.sectorsize))/fatconfig.clustersize;

		if(ScanFileAndAddToFAT(floppycontext,imgfile,0,&flatimg[fatposition],&flatimg[rootposition],&flatimg[dataposition],0,&fatconfig,numberofcluster))
		{
			return LOADER_BADFILE;
		}
		memcpy(&flatimg[((fatconfig.reservedsector)+(fatconfig.nbofsectorperfat))*fatconfig.sectorsize],&flatimg[fatposition],fatconfig.nbofsectorperfat*fatconfig.sectorsize);			


	}
	else
	{
		return LOADER_INTERNALERROR;
	}

	gap3len=84;
	interleave=1;		
	sectorsize=512;
	floppydisk->floppyiftype=IBMPC_HD_FLOPPYMODE;
	floppydisk->tracks=(CYLINDER**)malloc(sizeof(CYLINDER*)*floppydisk->floppyNumberOfTrack);
			
	tracklen=(DEFAULT_HD_BITRATE/(rpm/60))/4;
	trackdata=(unsigned char*)malloc(sectorsize*floppydisk->floppySectorPerTrack);
			
	for(j=0;j<floppydisk->floppyNumberOfTrack;j++)
	{
				
		floppydisk->tracks[j]=(CYLINDER*)malloc(sizeof(CYLINDER));
		currentcylinder=floppydisk->tracks[j];
		currentcylinder->number_of_side=floppydisk->floppyNumberOfSide;
		currentcylinder->sides=(SIDE**)malloc(sizeof(SIDE*)*currentcylinder->number_of_side);
		memset(currentcylinder->sides,0,sizeof(SIDE*)*currentcylinder->number_of_side);
				
		for(i=0;i<floppydisk->floppyNumberOfSide;i++)
		{
					
			currentcylinder->floppyRPM=rpm;
					
			currentcylinder->sides[i]=malloc(sizeof(SIDE));
			memset(currentcylinder->sides[i],0,sizeof(SIDE));
			currentside=currentcylinder->sides[i];		
			currentside->number_of_sector=floppydisk->floppySectorPerTrack;
			currentside->tracklen=tracklen;
					
			currentside->databuffer=malloc(currentside->tracklen);
			memset(currentside->databuffer,0,currentside->tracklen);
					
			currentside->flakybitsbuffer=0;

			currentside->timingbuffer=0;
			currentside->bitrate=DEFAULT_HD_BITRATE;

			currentside->indexbuffer=malloc(currentside->tracklen);
			memset(currentside->indexbuffer,0,currentside->tracklen);		
			fillindex(currentside->tracklen-1,currentside,2500,1);
			
			file_offset=(sectorsize*(j*currentside->number_of_sector*floppydisk->floppyNumberOfSide))+
						(sectorsize*(currentside->number_of_sector)*i);
					
			memcpy(trackdata,&flatimg[file_offset],sectorsize*currentside->number_of_sector);
					
					
			BuildISOTrack(floppycontext,IBMFORMAT_DD,currentside->number_of_sector,1,sectorsize,j,i,gap3len,trackdata,currentside->databuffer,&currentside->tracklen,interleave,0,NULL);
		}
	}
			
	free(flatimg);
			
	floppycontext->hxc_printf(MSG_INFO_1,"track file successfully loaded and encoded!");
	return LOADER_NOERROR;
}
	

